%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(gen_fsm_SUITE).

-include("test_server.hrl").

%% Test cases
-export([all/1]).

-export([start/1, start1/1, start2/1, start3/1, start4/1 , start5/1, start6/1,
	 start7/1, start8/1, start9/1, start10/1, start11/1]).

-export([abnormal/1, abnormal1/1, abnormal2/1]).

-export([sys/1, sys1/1]).

%% Exports for apply
-export([do_msg/1, do_sync_msg/1]).

% The gen_fsm behaviour
-export([init/1, handle_event/3, handle_sync_event/4, terminate/3]).
-export([idle/2,	idle/3,
	 wfor_conf/2,	wfor_conf/3,
	 connected/2,	connected/3]).
	 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


all(suite) -> {req,[stdlib],[start, abnormal, sys]}.


start(suite) -> [start1, start2, start3, start4, start5, start6, start7,
		 start8, start9, start10, start11].

%% anonymous
start1(suite) -> {req,{time,5}};
start1(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),

    ?line {ok, Pid0} = gen_fsm:start_link(gen_fsm_SUITE, [], []),
    ?line ok = do_func_test(Pid0),
    ?line ok = do_sync_func_test(Pid0),
    stop_it(Pid0),
%%    ?line stopped = gen_fsm:sync_send_all_state_event(Pid0, stop),
%%    ?line {'EXIT', {timeout,_}} = 
%%	(catch gen_fsm:sync_send_event(Pid0, hej)),

    ?line test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
   ok.

%% anonymous w. shutdown
start2(suite) -> {req,{time,5}};
start2(Config) when list(Config) ->
    %% Dont link when shutdown
    ?line {ok, Pid0} = gen_fsm:start(gen_fsm_SUITE, [], []),
    ?line ok = do_func_test(Pid0),
    ?line ok = do_sync_func_test(Pid0),
    ?line shutdown_stopped = 
	gen_fsm:sync_send_all_state_event(Pid0, stop_shutdown),
    ?line {'EXIT', {noproc,_}} = 
	(catch gen_fsm:sync_send_event(Pid0, hej)),

    ?line test_server:messages_get(),
    ok.

%% anonymous with timeout
start3(suite) -> {req,{time,5}};
start3(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),

    ?line {ok, Pid0} = gen_fsm:start(gen_fsm_SUITE, [], [{timeout,5}]),
    ?line ok = do_func_test(Pid0),
    ?line ok = do_sync_func_test(Pid0),
    ?line stop_it(Pid0),
    
    ?line {error, timeout} = gen_fsm:start(gen_fsm_SUITE, sleep,
					   [{timeout,5}]),

    test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
    ok.

%% anonymous with ignore
start4(suite) -> [];
start4(Config) when list(Config) ->
    OldFl = process_flag(trap_exit, true),

    ?line ignore = gen_fsm:start(gen_fsm_SUITE, ignore, []),

    test_server:messages_get(),
    process_flag(trap_exit, OldFl),
    ok.

%% anonymous with stop
start5(suite) -> [];
start5(Config) when list(Config) ->
    OldFl = process_flag(trap_exit, true),

    ?line {error, stopped} = gen_fsm:start(gen_fsm_SUITE, stop, []),

    test_server:messages_get(),
    process_flag(trap_exit, OldFl),
    ok.

%% anonymous linked
start6(suite) -> {req,{time,5}};
start6(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),

    ?line {ok, Pid} = gen_fsm:start_link(gen_fsm_SUITE, [], []),
    ?line ok = do_func_test(Pid),
    ?line ok = do_sync_func_test(Pid),
    ?line stop_it(Pid),
    
    test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
    ok.

%% global register linked
start7(suite) -> {req,{time,5}};
start7(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),
    
    ?line {ok, Pid} = 
	gen_fsm:start_link({global, my_fsm}, gen_fsm_SUITE, [], []),
    ?line {error, {already_started, Pid}} =
	gen_fsm:start_link({global, my_fsm}, gen_fsm_SUITE, [], []),
%%    ?line {error, {already_started, Pid}} =
%%	gen_fsm:start_link({local, my_fsm}, gen_fsm_SUITE, [], []),
    ?line {error, {already_started, Pid}} =
	gen_fsm:start({global, my_fsm}, gen_fsm_SUITE, [], []),
%%    ?line {error, {already_started, Pid}} =
%%	gen_fsm:start({local, my_fsm}, gen_fsm_SUITE, [], []),
    
    ?line ok = do_func_test(Pid),
    ?line ok = do_sync_func_test(Pid),
    ?line ok = do_func_test({global, my_fsm}),
    ?line ok = do_sync_func_test({global, my_fsm}),
    ?line stop_it({global, my_fsm}),
    
    test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
    ok.


%% local register
start8(suite) -> {req,{time,5}};
start8(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),

    ?line {ok, Pid} = 
	gen_fsm:start({local, my_fsm}, gen_fsm_SUITE, [], []),
    ?line {error, {already_started, Pid}} =
	gen_fsm:start({local, my_fsm}, gen_fsm_SUITE, [], []),

    ?line ok = do_func_test(Pid),
    ?line ok = do_sync_func_test(Pid),
    ?line ok = do_func_test(my_fsm),
    ?line ok = do_sync_func_test(my_fsm),
    ?line stop_it(Pid),
    
    test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
    ok.

%% local register linked
start9(suite) -> {req,{time,5}};
start9(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),

    ?line {ok, Pid} = 
	gen_fsm:start_link({local, my_fsm}, gen_fsm_SUITE, [], []),
    ?line {error, {already_started, Pid}} =
	gen_fsm:start({local, my_fsm}, gen_fsm_SUITE, [], []),

    ?line ok = do_func_test(Pid),
    ?line ok = do_sync_func_test(Pid),
    ?line ok = do_func_test(my_fsm),
    ?line ok = do_sync_func_test(my_fsm),
    ?line stop_it(Pid),
    
    test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
    ok.

%% global register
start10(suite) -> {req,{time,5}};
start10(Config) when list(Config) ->
    %%OldFl = process_flag(trap_exit, true),
    
    ?line {ok, Pid} = 
	gen_fsm:start({global, my_fsm}, gen_fsm_SUITE, [], []),
    ?line {error, {already_started, Pid}} =
	gen_fsm:start({global, my_fsm}, gen_fsm_SUITE, [], []),
%%    ?line {error, {already_started, Pid}} =
%%	gen_fsm:start({local, my_fsm}, gen_fsm_SUITE, [], []),
    ?line {error, {already_started, Pid}} =
	gen_fsm:start_link({global, my_fsm}, gen_fsm_SUITE, [], []),
%%    ?line {error, {already_started, Pid}} =
%%	gen_fsm:start_link({local, my_fsm}, gen_fsm_SUITE, [], []),
    
    ?line ok = do_func_test(Pid),
    ?line ok = do_sync_func_test(Pid),
    ?line ok = do_func_test({global, my_fsm}),
    ?line ok = do_sync_func_test({global, my_fsm}),
    ?line stop_it({global, my_fsm}),
    
    test_server:messages_get(),
    %%process_flag(trap_exit, OldFl),
    ok.


%% Stop registered processes
start11(suite) -> {req,{time,15}};
start11(Config) when list(Config) ->
    ?line {ok, Pid} = 
	gen_fsm:start_link({local, my_fsm}, gen_fsm_SUITE, [], []),
    ?line stop_it(Pid),

    ?line {ok, Pid1} = 
	gen_fsm:start_link({local, my_fsm}, gen_fsm_SUITE, [], []),
    ?line stop_it(my_fsm),
    
    ?line {ok, Pid2} = 
	gen_fsm:start({global, my_fsm}, gen_fsm_SUITE, [], []),
    ?line stop_it(Pid2),
    receive after 1 -> true end,
    ?line Result = 
	gen_fsm:start({global, my_fsm}, gen_fsm_SUITE, [], []),
    io:format("Result = ~p~n",[Result]),
    ?line {ok, Pid3} = Result, 
    ?line stop_it({global, my_fsm}),

    test_server:messages_get(),
    ok.

abnormal(suite) -> [abnormal1, abnormal2].

%% Check that time outs in calls work
abnormal1(suite) -> [];
abnormal1(Config) when list(Config) ->
    ?line {ok, Pid} = 
	gen_fsm:start({local, my_fsm}, gen_fsm_SUITE, [], []),

    %% timeout call.
    case os:type() of
	vxworks ->
	    %% timeout call for VxWorks must be in 16ms increments.
	    ?line delayed = gen_fsm:sync_send_event(my_fsm, {delayed_answer,1}, 17),
	    ?line {'EXIT',{timeout,_}} =
		(catch gen_fsm:sync_send_event(my_fsm, {delayed_answer,17}, 1));
	_ ->
	    ?line delayed = gen_fsm:sync_send_event(my_fsm, {delayed_answer,1}, 100),
	    ?line {'EXIT',{timeout,_}} =
		(catch gen_fsm:sync_send_event(my_fsm, {delayed_answer,10}, 1))
    end,
    test_server:messages_get(),
    ok.

%% Check that bad return values makes the fsm crash. Note that we must
%% trap exit since we must link to get the real bad_return_ error
abnormal2(suite) -> [];
abnormal2(Config) when list(Config) ->
    OldFl = process_flag(trap_exit, true),
    ?line {ok, Pid} = 
	gen_fsm:start_link(gen_fsm_SUITE, [], []),

    %% bad return value in the gen_fsm loop
    ?line {'EXIT',{{bad_return_value, badreturn},_}} =
	(catch gen_fsm:sync_send_event(Pid, badreturn)),
    
    test_server:messages_get(),
    process_flag(trap_exit, OldFl),
    ok.

sys(suite) -> [sys1].

sys1(suite) -> {req,{time,10}};
sys1(Config) when list(Config) ->
    ?line {ok, Pid} = 
	gen_fsm:start(gen_fsm_SUITE, [], []),
    ?line {status, Pid, {module,gen_fsm}, _} = sys:get_status(Pid),
    ?line sys:suspend(Pid),
    ?line {'EXIT', {timeout,_}} = 
	(catch gen_fsm:sync_send_event(Pid, hej)),
    ?line sys:resume(Pid),
    ?line stop_it(Pid).




%%sys1(suite) -> [];
%%sys1(_) ->




%%
%% Functionality check
%%

wfor(Msg) ->
    receive 
	Msg -> ok
    after 5000 -> 
	    throw(timeout)
    end.


stop_it(FSM) ->
    ?line stopped = 
	gen_fsm:sync_send_all_state_event(FSM, stop),
    if  pid(FSM) ->
	    ?line {'EXIT', {noproc,_}} = 
		(catch gen_fsm:sync_send_event(FSM, hej));
	true ->
	    ?line {'EXIT', _} = 
		(catch gen_fsm:sync_send_event(FSM, hej))
    end,
    ok.



do_func_test(FSM) ->
    ok = gen_fsm:send_all_state_event(FSM, {'alive?', self()}),
    wfor(yes),
    ok = do_connect(FSM),
    ok = gen_fsm:send_all_state_event(FSM, {'alive?', self()}),
    wfor(yes),
    test_server:do_times(3, ?MODULE, do_msg, [FSM]),
    ok = gen_fsm:send_all_state_event(FSM, {'alive?', self()}),
    wfor(yes),
    ok = do_disconnect(FSM),
    ok = gen_fsm:send_all_state_event(FSM, {'alive?', self()}),
    wfor(yes),
    ok.


do_connect(FSM) ->
    check_state(FSM, idle),
    gen_fsm:send_event(FSM, {connect, self()}),
    wfor(accept),
    check_state(FSM, wfor_conf),
    gen_fsm:send_event(FSM, confirmation),
    check_state(FSM, connected),
    ok.

do_msg(FSM) ->
    check_state(FSM, connected),
    R = make_ref(),
    ok = gen_fsm:send_event(FSM, {msg, R, self(), hej_pa_dig_quasimodo}),
    wfor({ak, R}).


do_disconnect(FSM) ->
    ok = gen_fsm:send_event(FSM, disconnect),
    check_state(FSM, idle).

check_state(FSM, State) ->
    case gen_fsm:sync_send_all_state_event(FSM, {get, self()}) of
	{state, State, _} -> ok
    end.

do_sync_func_test(FSM) ->
    yes = gen_fsm:sync_send_all_state_event(FSM, 'alive?'),
    ok = do_sync_connect(FSM),
    yes = gen_fsm:sync_send_all_state_event(FSM, 'alive?'),
    test_server:do_times(3, ?MODULE, do_sync_msg, [FSM]),
    yes = gen_fsm:sync_send_all_state_event(FSM, 'alive?'),
    ok = do_sync_disconnect(FSM),
    yes = gen_fsm:sync_send_all_state_event(FSM, 'alive?'),
    check_state(FSM, idle),
    ok.


do_sync_connect(FSM) ->
    check_state(FSM, idle),
    accept = gen_fsm:sync_send_event(FSM, {connect, self()}),
    check_state(FSM, wfor_conf),
    yes = gen_fsm:sync_send_event(FSM, confirmation),
    check_state(FSM, connected),
    ok.

do_sync_msg(FSM) ->
    check_state(FSM, connected),
    R = make_ref(),
    Res = gen_fsm:sync_send_event(FSM, {msg, R, self(), hej_pa_dig_quasimodo}),
    if  Res == {ak, R} ->
	    ok
    end.

do_sync_disconnect(FSM) ->
    yes = gen_fsm:sync_send_event(FSM, disconnect),
    check_state(FSM, idle).

    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% The Finite State Machine
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

init(ignore) ->
    ignore;
init(stop) ->
    {stop, stopped};
init(stop_shutdown) ->
    {stop, shutdown};
init(sleep) ->
    test_server:sleep(1000),
    {ok, idle, data};
init({timeout, T}) ->
    {ok, idle, state, T};
init(_) ->
    {ok, idle, state_data}.


terminate({From, stopped}, State, Data) ->
    From ! {self(), {stopped, State}},
    ok;
terminate(Reason, State, Data) ->
    ok.


idle({connect, Pid}, Data) ->
    Pid ! accept,
    {next_state, wfor_conf, Data};
idle(badreturn, Data) ->
    badreturn;
idle(_, Data) ->
    {next_state, idle, Data}.

idle({connect, Pid}, From, Data) ->
    {reply, accept, wfor_conf, Data};
idle({delayed_answer, T}, From, Data) ->
    test_server:sleep(T),
    {reply, delayed, idle, Data};
idle(badreturn, From, Data) ->
    badreturn;
idle(_, From, Data) ->
    {reply, 'eh?', idle, Data}.

wfor_conf(confirmation, Data) ->
    {next_state, connected, Data};
wfor_conf(_, Data) ->
    {next_state, idle, Data}.

wfor_conf(confirmation, From, Data) ->
    {reply, yes, connected, Data};
wfor_conf(_, From, Data) ->
    {reply, 'eh?', idle, Data}.

connected({msg, Ref, From, Msg}, Data) ->
    From ! {ak, Ref},
    {next_state, connected, Data};
connected(disconnect, Data) ->
    {next_state, idle, Data};
connected(_, Data) ->
    {next_state, connected, Data}.

connected({msg, Ref, From, Msg}, _, Data) ->
    {reply, {ak, Ref}, connected, Data};
connected(disconnect, From, Data) ->
    {reply, yes, idle, Data};
connected(_, _, Data) ->
    {reply, 'eh?', connected, Data}.

handle_event({get, Pid}, State, Data) ->
    Pid ! {state, State, Data},
    {next_state, State, Data};
handle_event(stop, State, Data) ->
    {stop, normal, Data};
handle_event(stop_shutdown, State, Data) ->
    {stop, shutdown, Data};
handle_event({'alive?', Pid}, State, Data) ->
    Pid ! yes,
    {next_state, State, Data}.

handle_sync_event('alive?', From, State, Data) ->
    {reply, yes, State, Data};
handle_sync_event(stop, From, State, Data) ->
    {stop, normal, stopped, Data};
handle_sync_event(stop_shutdown, From, State, Data) ->
    {stop, shutdown, shutdown_stopped, Data};
handle_sync_event({get, Pid}, From, State, Data) ->
    {reply, {state, State, Data}, State, Data}.
    
    










